@@ -12,27 +12,31 @@ class Database:
     SUM = 'SUM'
     POW = 'POW'
 
-    TABLE_PARAMS = "params"
-    TABLE_CORPUS = "corpus"
-    TABLE_WEIGHTS = "weights"
+    def __init__(self, engine: Engine, prefix, type_features, type_classes,
+                 field_features, field_classes, field_items, field_weights,
+                 table_params, table_corpus, table_weights):
 
-    n = FIELD_ITEM = "item"
-    j = FIELD_FEATURE = "feature"
-    k = FIELD_CLASS = "class"
-    w = FIELD_WEIGHT = "weight"
-
-    def __init__(self, engine: Engine, prefix, type_features, type_classes):
         self.prefix = prefix
+
         self.type_features = type_features
         self.type_classes = type_classes
 
+        self.table_params = table_params
+        self.table_corpus = table_corpus
+        self.table_weights = table_weights
+
+        self.j = self.field_features = field_features
+        self.k = self.field_classes = field_classes
+        self.n = self.field_items = field_items
+        self.w = self.field_weights = field_weights
+
         if isinstance(engine, str):
             self.engine = create_engine(engine, echo=False)
         else:
             self.engine = engine
 
         self.table_params = Table(
-            f"{self.prefix}_{self.TABLE_PARAMS}", MetaData(),
+            f"{self.prefix}_{self.table_params}", MetaData(),
             Column('prefix', String, primary_key=True),
             Column('a', Float),
             Column('b', Float),
@@ -40,22 +44,23 @@ class Database:
         )
 
         self.table_corpus = Table(
-            f"{self.prefix}_{self.TABLE_CORPUS}", MetaData(),
-            Column(self.FIELD_FEATURE, self.type_features, primary_key=True),
-            Column(self.FIELD_CLASS, self.type_classes, primary_key=True),
-            Column(self.FIELD_WEIGHT, Float),
+            f"{self.prefix}_{self.table_corpus}", MetaData(),
+            Column(self.field_features, self.type_features, primary_key=True),
+            Column(self.field_classes, self.type_classes, primary_key=True),
+            Column(self.field_weights, Float),
         )
 
         self.table_weights = Table(
-            f"{self.prefix}_{self.TABLE_WEIGHTS}", MetaData(),
-            Column(self.FIELD_FEATURE, self.type_features, primary_key=True),
-            Column(self.FIELD_CLASS, self.type_classes, primary_key=True),
-            Column(self.FIELD_WEIGHT, Float),
+            f"{self.prefix}_{self.table_weights}", MetaData(),
+            Column(self.field_features, self.type_features, primary_key=True),
+            Column(self.field_classes, self.type_classes, primary_key=True),
+            Column(self.field_weights, Float),
         )
 
         if not self.exists(self.table_params):
             with self.connect() as con:
-                self.write_params(con, a=0.5, b=1, h=1)
+                with con.begin():
+                    self.write_params(con, a=0.5, b=1, h=1)
 
     def connect(self):
         return self.engine.connect()
@@ -137,12 +142,19 @@ class Database:
         return con.execute(text(sql), values)
 
     def read_params(self, con):
-        params = con.execute(f"SELECT a, b, h FROM {self.table_params}")
+        params = con.execute(text(f"SELECT a, b, h FROM {self.table_params}"))
         values = params.fetchone()
         keys = params.keys()
 
         return dict(zip(keys, values))
 
+    def read_sql(self, sql, con):
+        cur = con.execute(text(sql) if isinstance(sql, str) else sql)
+        data = [tuple(row) for row in cur.fetchall()]
+        columns = cur.keys()
+
+        return pd.DataFrame(data, columns=columns)
+
     def write_params(self, con, a, b, h):
         if_exists = {
             'if_exists': 'insert_or_replace',
@@ -156,8 +168,8 @@ class Database:
     def write_corpus(self, con, X, y, sample_weight):
         if_exists = {
             'if_exists': 'insert_or_sum',
-            'conflict': [self.FIELD_CLASS, self.FIELD_FEATURE],
-            'sum': [self.FIELD_WEIGHT]
+            'conflict': [self.field_classes, self.field_features],
+            'sum': [self.field_weights]
         }
 
         corpus = defaultdict(lambda: defaultdict(int))
@@ -173,21 +185,21 @@ class Database:
         values = []
         for c, d in corpus.items():
             for f, w in d.items():
-                values.append({self.FIELD_FEATURE: f, self.FIELD_CLASS: c, self.FIELD_WEIGHT: w})
+                values.append({self.field_features: f, self.field_classes: c, self.field_weights: w})
 
         return self.write(con, table=self.table_corpus, values=values, **if_exists)
 
     def write_items(self, con, X):
         table = Table(
             f"{self.prefix}_items_{md5(str(uuid1()).encode()).hexdigest()[:12]}", MetaData(),
-            Column(self.FIELD_ITEM, Integer, primary_key=True),
-            Column(self.FIELD_FEATURE, self.type_features, primary_key=True),
-            Column(self.FIELD_WEIGHT, Float),
+            Column(self.field_items, Integer, primary_key=True),
+            Column(self.field_features, self.type_features, primary_key=True),
+            Column(self.field_weights, Float),
             prefixes=["TEMPORARY"],
         )
 
         values = [
-            {self.FIELD_ITEM: i, self.FIELD_FEATURE: f, self.FIELD_WEIGHT: w}
+            {self.field_items: i, self.field_features: f, self.field_weights: w}
             for i, x in enumerate(X)
             for f, w in x.items()
         ]
@@ -197,35 +209,53 @@ class Database:
     def write_sample_weight(self, con, sample_weight):
         table = Table(
             f"{self.prefix}_sample_weight_{md5(str(uuid1()).encode()).hexdigest()[:12]}", MetaData(),
-            Column(self.FIELD_ITEM, Integer, primary_key=True),
-            Column(self.FIELD_WEIGHT, Float),
+            Column(self.field_items, Integer, primary_key=True),
+            Column(self.field_weights, Float),
             prefixes=["TEMPORARY"],
         )
 
         values = [
-            {self.FIELD_ITEM: i, self.FIELD_WEIGHT: w}
+            {self.field_items: i, self.field_weights: w}
             for i, w in enumerate(sample_weight)
         ]
 
         return self.write(con, table=table, values=values)
 
+    def is_deployed(self):
+        return self.exists(self.table_weights)
+
+    def is_fitted(self):
+        return self.exists(self.table_corpus) or self.is_deployed()
+
+    def check_fitted(self):
+        if not self.is_fitted():
+            raise ValueError(
+                f"This instance is not fitted yet."
+            )
+
+    def check_editable(self):
+        if self.is_deployed():
+            raise ValueError(
+                "Cannot modify a deployed instance."
+            )
+
     def deploy(self, con):
         self.check_fitted()
 
-        if self.exists(self.table_weights):
+        if self.is_deployed():
             raise ValueError(
                 "This instance is already deployed. Nothing to do."
             )
 
         sql = f"""
             INSERT INTO  {self.table_weights} 
-            {self._sql_WITH()} 
+            {self._sql_WITH(cache=False)} 
             SELECT {self.j}, {self.k}, {self.w} 
             FROM HW_jk
             """
 
         self.write(con, table=self.table_weights)
-        con.execute(sql)
+        con.execute(text(sql))
 
     def undeploy(self, con):
         if not self.exists(self.table_corpus):
@@ -233,40 +263,32 @@ class Database:
                 "This instance has no corpus. Cannot undeploy: information would be lost."
             )
 
-        if not self.exists(self.table_weights):
+        if not self.is_deployed():
             raise ValueError(
                 "This instance is already undeployed. Nothing to do."
             )
 
         self.table_weights.drop(con)
 
-    def check_fitted(self):
-        if not self.exists(self.table_weights) and not self.exists(self.table_corpus):
-            raise ValueError(
-                f"This instance is not fitted yet."
-            )
-
-    def check_editable(self):
-        if self.exists(self.table_weights):
-            raise ValueError(
-                "Cannot modify a deployed instance."
-            )
-
     def predict(self, con, X):
+        cache = self.is_deployed()
         items = self.write_items(con, X)
-        sql = self._sql_predict(items)
+        sql = self._sql_predict(cache, items)
 
-        return pd.read_sql(sql, con)
+        return self.read_sql(sql, con)
 
     def predict_proba(self, con, X):
+        cache = self.is_deployed()
         items = self.write_items(con, X)
-        sql = self._sql_predict_proba(items)
+        sql = self._sql_predict_proba(cache, items)
 
-        return pd.read_sql(sql, con)
+        return self.read_sql(sql, con)
 
     def explain(self, con, X=None, sample_weight=None):
+        cache = self.is_deployed()
+
         if X is None:
-            sql = self._sql_explain()
+            sql = self._sql_explain(cache)
 
         else:
             norm = [sum(v for k, v in x.items()) for x in X]
@@ -278,13 +300,13 @@ class Database:
                 X = [{k: pow(w, p) * v / n for k, v in x.items()} for x, n, w in zip(X, norm, sample_weight) if n > 0]
 
             items = self.write_items(con, X)
-            sql = self._sql_explain(items)
+            sql = self._sql_explain(cache, items)
 
-        return pd.read_sql(sql, con)
+        return self.read_sql(sql, con)
 
-    def _sql_predict(self, items):
+    def _sql_predict(self, cache, items):
         return f"""
-            {self._sql_WITH()}, 
+            {self._sql_WITH(cache)}, 
                 X_njk AS ({self._sql_X_njk(items)}), 
                 X_nk AS ({self._sql_X_nk()}), 
                 R_nk AS (
@@ -304,9 +326,9 @@ class Database:
                 idx = 1
             """
 
-    def _sql_predict_proba(self, items):
+    def _sql_predict_proba(self, cache, items):
         return f"""
-            {self._sql_WITH()}, 
+            {self._sql_WITH(cache)}, 
                 X_njk AS ({self._sql_X_njk(items)}), 
                 X_nk AS ({self._sql_X_nk()}), 
                 U_nk AS ({self._sql_U_nk()}), 
@@ -320,10 +342,10 @@ class Database:
                 Y_nk
             """
 
-    def _sql_explain(self, items=None):
+    def _sql_explain(self, cache, items=None):
         if items is None:
             return f"""
-                {self._sql_WITH()} 
+                {self._sql_WITH(cache)} 
                 SELECT 
                     {self.j},
                     {self.k},
@@ -333,7 +355,7 @@ class Database:
                 """
 
         return f"""
-            {self._sql_WITH()}, 
+            {self._sql_WITH(cache)}, 
                 X_njk AS ({self._sql_X_njk(items)})
             SELECT 
                 {self.j},
@@ -347,7 +369,7 @@ class Database:
             """
 
         # return f"""
-        #     {self._sql_WITH()},
+        #     {self._sql_WITH(cache)},
         #         X_njk AS ({self._sql_X_njk(items)}),
         #         X_nk AS ({self._sql_X_nk()}),
         #         U_nk AS ({self._sql_U_nk()}),
@@ -374,8 +396,8 @@ class Database:
         #         Y_njk.{self.k}
         #     """
 
-    def _sql_WITH(self):
-        if self.exists(self.table_weights):
+    def _sql_WITH(self, cache):
+        if cache:
             sql = f"""
                 WITH 
                     VAL AS ({self._sql_VAL()}),
